/*  This file is part of kbdlock
 *  Copyright (C) 1999,2005 Sergey Poznyakoff
 *
 *  Kbdlock is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Kbdlock is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with GNU cflow; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */
#include <getopt.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/vt.h>
#include <pwd.h>
#include <signal.h>
#include <crypt.h>
#ifdef __SHADOW__
# include <shadow.h>
#endif

char rccsid[] = "$Id: main.c,v 1.4 2005/04/04 16:36:40 gray Exp $";

char *device = "/dev/tty";

#ifdef GNU_STYLE_OPTIONS
typedef struct option LONGOPT;
/* Note: Internal variables go first */
LONGOPT longopts[] = {
    "help",    no_argument, 0, 'h',
    "license", no_argument, 0, 'l',
    "password",   no_argument, 0, 'p',
    "timeout",    required_argument, 0, 't',
    "version", no_argument, 0, 'v',
    "quiet",   no_argument, 0, 'q',
    0,
};
#else
#define getopt_long(argc, argv, optstr, long_opts, iptr) \
 getopt(argc, argv, optstr)
#endif

char *progname;
char usage[] =
"usage: kbdlock [-hlptvq]\n"
"options are:\n"
"      -p, --password              prompt for password\n"
"      -t TO, --timeout TO         set timeout to TO.\n"
"      -q, --quiet                 be quiet\n"
"Following options cause kbdlock to exit with zero status:\n"
"      -h, --help                  display this help screen\n"
"      -l, --license               display license\n"
"      -v, --version               display version\n";

char licence[] = "licence\n";
int quiet = 0;
int use_timeout = 0;
long timeout;

static long gettimeout(char *str);
static int checktty();

main(argc, argv)
    int argc;
    char **argv;
{
    int fd, c, rc;
    fd_set rfd;
    struct timeval tv;
    uid_t uid = getuid();
    struct passwd *pwd;
#ifdef __SHADOW__
    struct spwd *spwd;
#endif    
    char *p1, *p2, *username;
    char *pass;
    char salt[3];

    progname = argv[0];

    pwd = getpwuid(uid);
    username = pwd->pw_name;
#ifdef __SHADOW__
    spwd = getspnam(username);
    if (!spwd) {
	printf("%s has wrong permissions\n", progname);
	exit(1);
    }
    pass = strdup(spwd->sp_pwdp);
#else
    if (!pwd->pw_passwd[0] || strlen(pwd->pw_passwd) == 1) {
	printf("%s has wrong permissions\n", progname);
	exit(1);
    }
    pass = pwd->pw_passwd;
#endif

    while ((c = getopt_long(argc, argv, "hlpt:vq", longopts, &rc)) != EOF) {
	switch (c) {
	case 'h':
	    printf("%s", usage);
	    exit(0);
	case 'l':
	    printf("%s", licence);
	    exit(0);
	case 'p':
	    p1 = strdup(getpass("password:"));
	    p2 = strdup(getpass("re-enter:"));
	    if (strcmp(p1, p2)) {
		printf("entries differ");
		exit(1);
	    }
	    memset(p1, 0, strlen(p1));
	    free(p1);

	    srand(time(NULL));
	    strcpy(salt, "$1");
	    pass = strdup(crypt(p2, salt));

            memset(p2, 0, strlen(p2));
	    free(p2);
	    break;
	case 't':
	    use_timeout++;
	    timeout = gettimeout(optarg);
	    break;
	case 'v':
	    printf("%s\n", rccsid);
	    exit(0);
	case 'q':
	    quiet++;
	    break;
	default:
	    printf("%s", usage);
	    exit(1);
	}
    }

    if (optind < argc) {
	printf("%s: unexpected argument\n", progname);
	exit(1);
    }
    
    checktty();

    setvbuf(stdout, NULL, _IONBF, 0);
	
#if !defined(__DEBUG__)
    signal(SIGHUP, SIG_IGN);
    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGILL, SIG_IGN);
    signal(SIGTRAP, SIG_IGN);
    signal(SIGABRT, SIG_IGN);
    signal(SIGBUS, SIG_IGN);
    signal(SIGFPE, SIG_IGN);
    signal(SIGKILL, SIG_IGN);
    signal(SIGUSR1, SIG_IGN);
    signal(SIGSEGV, SIG_IGN);
    signal(SIGUSR2, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGALRM, SIG_IGN);
    signal(SIGTERM, SIG_IGN);
    signal(SIGCHLD, SIG_IGN);
    signal(SIGCONT, SIG_IGN);
    signal(SIGSTOP, SIG_IGN);
    signal(SIGTSTP, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGTTOU, SIG_IGN);
    signal(SIGURG, SIG_IGN);
    signal(SIGXCPU, SIG_IGN);
    signal(SIGXFSZ, SIG_IGN);
    signal(SIGVTALRM, SIG_IGN);
    signal(SIGPROF, SIG_IGN);
    signal(SIGWINCH, SIG_IGN);
    signal(SIGIO, SIG_IGN);
    signal(SIGPWR, SIG_IGN);
    
    fd = open(device, O_RDONLY);
    if (fd == -1) {
	perror(device);
	exit(1);
    }
    rc = ioctl(fd, VT_LOCKSWITCH);
    /*    printf("ioctl: %d\n", rc);*/
#endif

    if (!quiet) {
	printf("tty is locked by %s", username);
	if (use_timeout)
	    printf(", timeout %ld sec.", timeout);
	printf("\n");
    }
    
    for (;;) {
	FD_ZERO(&rfd);
	FD_SET(0, &rfd);

	tv.tv_sec = 5;
	tv.tv_usec = 0;
	if (use_timeout && (timeout -= tv.tv_sec) <= 0)
	    break;
	    
	if (select(1, &rfd, NULL, NULL, &tv) < 0) {
	    continue;
	}

	if (FD_ISSET(0, &rfd)) {
	    read(1, &c, 1);
	    p1 = getpass("password:");
	    if (strcmp(crypt(p1, pass), pass) == 0)
		break;
	    printf("sorry\n");
	}
    }
#if !defined(__DEBUG__)
    ioctl(fd, VT_UNLOCKSWITCH);
    close(fd);
#endif
}

long
gettimeout(str)
    char *str;
{
    char *p;
    int n;
    long hours = 0, minutes = 0, seconds = 0;
    
    p = str;

    for (;;) {
	n = 0;
	while (*p && isdigit(*p)) {
	    n = 10*n + *p - '0';
	    p++;
	}
    
	switch (*p) {
	case 'h':
	case 'H':
	    if (hours) {
		printf("hours already specified\n");
		exit(1);
	    }
	    hours = n;
	    break;
	case 'm':
	case 'M':
	    if (minutes) {
		printf("minutes already specified\n");
		exit(1);
	    }
	    minutes = n;
	    break;
	case 's':
	case 'S':
	    if (seconds) {
		printf("seconds already specified\n");
		exit(1);
	    }
	    seconds = n;
	    break;
	case 0:
	    if (!hours && !minutes && !seconds)
		return n;
	    return (hours*60 + minutes)*60 + seconds;
	default:
	    printf("parse error near `%c'", *p);
	    exit(1);
	}
	p++;
    }
}


int
checktty()
{
    char *p;

    if (!isatty(0)) {
    	printf("%s: stdin not a terminal. Exiting...\n", progname);
    	exit(1);
    }

    p = ttyname(0);
    if (*p == '/')
	p++;
    for (; *p && *p != '/'; p++)
	;
    if (!*p) {
	printf("%s: hmm, weird terminal. I'd better leave...\n", progname);
	exit(1);
    }
    ++p;
    if (strncmp(p, "tty", 3) == 0)
	return !isdigit(p[3]);
    else if (strcmp(p, "console") == 0)
        return 0;
    return 1;
} 


